from collections import namedtuple
from singer.catalog import Catalog, CatalogEntry
from singer import metadata
from ..common import get_cursor
from . import schemas

Table = namedtuple("Table", [
    "table_schema",
    "table_name",
    "table_type",
])

Column = namedtuple("Column", [
    "table_schema",
    "table_name",
    "column_name",
    "data_type",
    "character_maximum_length",
    "numeric_precision",
    "numeric_scale",
])

SUPPORTED_TYPES = {"T", "V", "P"}


# Note the _query_* functions mainly exist for the sake of mocking in unit
# tests. Normally I would prefer to have integration tests than mock out this
# data, but DB2 databases aren't easy to come by and if there is a lot of data
# these queries can be quite slow.
def _query_tables(config):
    """Queries the qsys2 tables catalog and returns an iterator containing the
    raw results."""
    with get_cursor(config) as cursor:
        cursor.execute("""
            SELECT table_schema,
                   table_name,
                   table_type
              FROM qsys2.systables
             WHERE table_type IN ('T', 'V', 'P')
        """)
        yield from cursor


def _query_columns(config):
    """Queries the qsys2 columns catalog and returns an iterator containing the
    raw results."""
    with get_cursor(config) as cursor:
        cursor.execute("""
            SELECT table_schema,
                   table_name,
                   column_name,
                   data_type,
                   character_maximum_length,
                   numeric_precision,
                   numeric_scale
              FROM qsys2.syscolumns
        """)
        yield from cursor


def _query_primary_keys(config):
    """Queries the qsys2 primary key catalog and returns an iterator containing
    the raw results."""
    with get_cursor(config) as cursor:
        cursor.execute("""
            SELECT A.table_schema,
                   A.table_name,
                   A.column_name,
                   A.ordinal_position
              FROM qsys2.syskeycst A
              JOIN qsys2.syscst B
                ON A.constraint_schema = B.constraint_schema
               AND A.constraint_name = B.constraint_name
             WHERE B.constraint_type = 'PRIMARY KEY'
        """)
        yield from cursor


def _table_id(table):
    """Returns a 2-tuple that can be used to uniquely identify the table."""
    return (table.table_schema, table.table_name)


def _find_tables(config):
    # We use the qsys2.systables catalog rather than
    # information_schema.tables because it contains better information
    # about the "table_type." The information_schema table doesn't
    # distinguish between tables and data files.
    tbls = [Table(*rec) for rec in _query_tables(config)]
    return {_table_id(t): t for t in tbls}


def _col_table_id(column):
    """Returns a 2-tuple representing the column's table ID. See also
    _table_id."""
    return (column.table_schema, column.table_name)


def _find_columns(config, tables):
    cols = (Column(*rec) for rec in _query_columns(config))
    ret = {}
    for col in cols:
        table_id = _col_table_id(col)
        if table_id not in tables:
            continue
        if tables[table_id].table_type not in SUPPORTED_TYPES:
            continue
        if table_id not in ret:
            ret[table_id] = []
        ret[table_id].append(col)
    return ret


def _find_primary_keys(config, tables):
    """Returns a dict of tuples -> list where the keys are \"table ids\" -
    ie. the (schema name, table_name) and the values are the primary key
    columns, sorted by their ordinal position."""
    results = _query_primary_keys(config)
    keys = {}
    for (table_schema, table_name, column_name, ordinal_pos) in results:
        table_id = (table_schema, table_name)
        if table_id not in tables:
            continue
        if table_id not in keys:
            keys[table_id] = []
        # We append a 2-tuple containing the ordinal position first so we
        # can sort the PKs by their ordinal position once this loop is
        # done.
        keys[table_id].append((ordinal_pos, column_name))
    return {
        k: [x[1] for x in sorted(v)]
        for k, v in keys.items()
    }


def _create_column_metadata(cols, schema):
    mdata = {}
    mdata = metadata.write(mdata, (), "selected-by-default", False)
    for col in cols:
        col_schema = schema.properties[col.column_name]
        mdata = metadata.write(mdata,
                               breadcrumb=("properties", col.column_name),
                               k="selected-by-default",
                               val=(col_schema.inclusion != "unsupported"))
        mdata = metadata.write(mdata,
                               breadcrumb=("properties", col.column_name),
                               k="sql-datatype",
                               val=col.data_type.lower())
    return metadata.to_list(mdata)


def _update_entry_for_table_type(catalog_entry, table_type):
    catalog_entry.is_view = table_type == "V"
    # https://www.ibm.com/support/knowledgecenter/ssw_ibm_i_71/db2/rbafzcatsystbls.htm
    known_types = {
        "A": "Alias",
        "L": "Logical file",
        "M": "Materialized query table",
        "P": "Physical file",
    }
    if table_type not in SUPPORTED_TYPES:
        catalog_entry.schema.inclusion = "unsupported"
        hint = known_types.get(table_type, "Unknown")
        err = "Unsupported table type {} ({})".format(table_type, hint)
        catalog_entry.schema.description = err


def discover(config):
    tables = _find_tables(config)
    columns = _find_columns(config, tables)
    pks = _find_primary_keys(config, tables)
    entries = []
    for table_id in tables:
        table_schema, table_name = table_id
        cols = columns.get(table_id, [])
        pk_columns = pks.get(table_id, [])
        schema = schemas.generate(cols, pk_columns)
        entry = CatalogEntry(
            database=table_schema,
            table=table_name,
            stream=table_name,
            metadata=_create_column_metadata(cols, schema),
            tap_stream_id="{}-{}".format(table_schema, table_name),
            schema=schema)
        if table_id in pks:
            entry.key_properties = pks[table_id]
        _update_entry_for_table_type(entry, tables[table_id].table_type)
        entries.append(entry)
    return Catalog(entries)



"""Functions which accept user input, like a state or catalog, and modifies
them for use during the sync process.

The functions herein were taken nearly verbatim from
https://github.com/singer-io/tap-mysql/"""
from itertools import dropwhile
import singer
import singer.schema
from singer.schema import Schema
from singer.catalog import Catalog, CatalogEntry

LOGGER = singer.get_logger()


def _desired_columns(selected, table_schema):

    '''Return the set of column names we need to include in the SELECT.
    selected - set of column names marked as selected in the input catalog
    table_schema - the most recently discovered Schema for the table
    '''
    all_columns = set()
    available = set()
    automatic = set()
    unsupported = set()

    for column, column_schema in table_schema.properties.items():
        all_columns.add(column)
        inclusion = column_schema.inclusion
        if inclusion == 'automatic':
            automatic.add(column)
        elif inclusion == 'available':
            available.add(column)
        elif inclusion == 'unsupported':
            unsupported.add(column)
        else:
            raise Exception('Unknown inclusion ' + inclusion)

    selected_but_unsupported = selected.intersection(unsupported)
    if selected_but_unsupported:
        LOGGER.warning(
            'Columns %s were selected but are not supported. Skipping them.',
            selected_but_unsupported)

    selected_but_nonexistent = selected.difference(all_columns)
    if selected_but_nonexistent:
        LOGGER.warning(
            'Columns %s were selected but do not exist.',
            selected_but_nonexistent)

    not_selected_but_automatic = automatic.difference(selected)
    if not_selected_but_automatic:
        LOGGER.warning(
            'Columns %s are primary keys but were not selected. Adding them.',
            not_selected_but_automatic)

    return selected.intersection(available).union(automatic)


def resolve_catalog(catalog, discovered, state):
    '''Returns the Catalog of data we're going to sync.
    Takes the Catalog we read from the input file and turns it into a
    Catalog representing exactly which tables and columns we're going to
    emit in this process. Compares the input Catalog to a freshly
    discovered Catalog to determine the resulting Catalog. Returns a new
    instance. The result may differ from the input Catalog in the
    following ways:
      * It will only include streams marked as "selected".
      * We will remove any streams and columns that were selected but do
        not actually exist in the database right now.
      * If the state has a currently_syncing, we will skip to that stream and
        drop all streams appearing before it in the catalog.
      * We will add any columns that were not selected but should be
        automatically included. For example, primary key columns and
        columns used as replication keys.
    '''

    # Filter catalog to include only selected streams
    streams = list(filter(lambda stream: stream.is_selected(), catalog.streams))

    # If the state says we were in the middle of processing a stream, skip
    # to that stream.
    currently_syncing = singer.get_currently_syncing(state)
    if currently_syncing:
        streams = dropwhile(lambda s: s.tap_stream_id != currently_syncing, streams)

    result = Catalog(streams=[])

    # Iterate over the streams in the input catalog and match each one up
    # with the same stream in the discovered catalog.
    for catalog_entry in streams:

        discovered_table = discovered.get_stream(catalog_entry.tap_stream_id)
        if not discovered_table:
            LOGGER.warning('Database %s table %s was selected but does not exist',
                           catalog_entry.database, catalog_entry.table)
            continue
        selected = set([k for k, v in catalog_entry.schema.properties.items()
                        if v.selected or k == catalog_entry.replication_key])

        # These are the columns we need to select
        columns = _desired_columns(selected, discovered_table.schema)

        result.streams.append(CatalogEntry(
            tap_stream_id=catalog_entry.tap_stream_id,
            key_properties=catalog_entry.key_properties,
            stream=catalog_entry.stream,
            metadata=catalog_entry.metadata,
            database=catalog_entry.database,
            table=catalog_entry.table,
            replication_key=catalog_entry.replication_key,
            is_view=catalog_entry.is_view,
            schema=Schema(
                type='object',
                properties={col: discovered_table.schema.properties[col]
                            for col in columns}
            )
        ))

    return result


def build_state(raw_state, catalog):
    state = {}

    currently_syncing = singer.get_currently_syncing(raw_state)
    if currently_syncing:
        state = singer.set_currently_syncing(state, currently_syncing)

    for catalog_entry in catalog.streams:
        if catalog_entry.replication_key:
            state = singer.write_bookmark(state,
                                          catalog_entry.tap_stream_id,
                                          'replication_key',
                                          catalog_entry.replication_key)

            # Only keep the existing replication_key_value if the
            # replication_key hasn't changed.
            raw_replication_key = singer.get_bookmark(raw_state,
                                                      catalog_entry.tap_stream_id,
                                                      'replication_key')
            if raw_replication_key == catalog_entry.replication_key:
                raw_replication_key_value = singer.get_bookmark(raw_state,
                                                                catalog_entry.tap_stream_id,
                                                                'replication_key_value')
                state = singer.write_bookmark(state,
                                              catalog_entry.tap_stream_id,
                                              'replication_key_value',
                                              raw_replication_key_value)

        # Persist any existing version, even if it's None
        if raw_state.get('bookmarks', {}).get(catalog_entry.tap_stream_id):
            raw_stream_version = singer.get_bookmark(raw_state,
                                                     catalog_entry.tap_stream_id,
                                                     'version')

            state = singer.write_bookmark(state,
                                          catalog_entry.tap_stream_id,
                                          'version',
                                          raw_stream_version)

    return state



from datetime import datetime, date, time
from time import time as time_
from collections import namedtuple
import pendulum
import singer
import singer.metrics as metrics
from singer.catalog import CatalogEntry
from singer import write_message as _emit
from singer import metadata
from .common import get_cursor

LOGGER = singer.get_logger()
_get_bk = singer.get_bookmark
_set_bk = singer.write_bookmark

ReplicationKey = namedtuple("ReplicationKey", ["column", "value"])


def _quote(x):
    return '"{}"'.format(x.replace('"', '""'))


def _get_stream_version(tap_stream_id, state):
    return _get_bk(state, tap_stream_id, "version") or int(time_() * 1000)


def _is_datetime_col(catalog_entry, column):
    return catalog_entry.schema.properties[column].format == "date-time"


def _get_replication_key(state: dict, catalog_entry: CatalogEntry):
    tap_stream_id = catalog_entry.tap_stream_id
    column = _get_bk(state, tap_stream_id, "replication_key")
    if not column:
        return None
    value = _get_bk(state, tap_stream_id, "replication_key_value")
    if value and _is_datetime_col(catalog_entry, column):
        value = pendulum.parse(value)
    return ReplicationKey(column, value)


def _column_sql(catalog_entry: CatalogEntry, column: str):
    data_type = metadata.get(metadata.to_map(catalog_entry.metadata),
                             breadcrumb=("properties", column),
                             k="sql-datatype")
    if data_type == "timestmp":  # No that's not a typo
        return "{} - CURRENT TIMEZONE".format(_quote(column))
    return _quote(column)


def _create_sql(catalog_entry: CatalogEntry, columns, rep_key: ReplicationKey):
    escaped_columns = [_column_sql(catalog_entry, c) for c in columns]
    select = "SELECT {} FROM {}.{}".format(
        ",".join(escaped_columns),
        _quote(catalog_entry.database),
        _quote(catalog_entry.table))
    params = ()
    if not rep_key:
        return select, params
    if rep_key.value:
        select += " WHERE {} >= ?".format(_quote(rep_key.column))
        params = (rep_key.value,)
    select += " ORDER BY {} ASC".format(_quote(rep_key.column))
    return select, params


def _row_to_record(catalog_entry, version, row, columns):
    row_to_persist = ()
    for elem in row:
        if isinstance(elem, datetime):
            row_to_persist += (elem.isoformat() + "+00:00",)
        elif isinstance(elem, date):
            row_to_persist += (elem.isoformat() + "T00:00:00+00:00",)
        elif isinstance(elem, time):
            epoch = datetime.utcfromtimestamp(0)
            epoch_with_time = datetime.combine(epoch.date(), elem)
            row_to_persist += (epoch_with_time.isoformat() + "+00:00",)
        else:
            row_to_persist += (elem,)
    return singer.RecordMessage(
        stream=catalog_entry.stream,
        record=dict(zip(columns, row_to_persist)),
        version=version)


def _write_metrics(catalog_entry, rows_saved):
    with metrics.record_counter(None) as counter:
        counter.tags["database"] = catalog_entry.database
        counter.tags["table"] = catalog_entry.table
        counter.increment(rows_saved)


def _activate_version(catalog_entry: CatalogEntry, version):
    _emit(singer.ActivateVersionMessage(stream=catalog_entry.stream,
                                        version=version))


def _maybe_activate_before_sync(state, catalog_entry, rep_key, stream_version):
    # If there's a replication key, we want to emit an ACTIVATE_VERSION
    # message at the beginning so the records show up right away. If
    # there's no bookmark at all for this stream, assume it's the very
    # first replication. That is, clients have never seen rows for this
    # stream before, so they can immediately acknowledge the present
    # version.
    tap_stream_id = catalog_entry.tap_stream_id
    bookmark_is_empty = not state.get("bookmarks", {}).get(tap_stream_id)
    if rep_key or bookmark_is_empty:
        _activate_version(catalog_entry, stream_version)


def _maybe_activate_after_sync(state, catalog_entry, rep_key, stream_version):
    # If there is no replication key, we"re doing "full table" replication,
    # and we need to activate this version at the end. Also clear the
    # stream's version from the state so that subsequent invocations will
    # emit a distinct stream version.
    if not rep_key:
        _activate_version(catalog_entry, stream_version)
        tap_stream_id = catalog_entry.tap_stream_id
        state = _set_bk(state, tap_stream_id, "version", None)
    return state


def _sync_table(config, state, catalog_entry):
    columns = list(catalog_entry.schema.properties)
    if not columns:
        LOGGER.warning(
            "There are no columns selected for table %s, skipping it",
            catalog_entry.table)
        return
    tap_stream_id = catalog_entry.tap_stream_id
    rep_key = _get_replication_key(state, catalog_entry)
    stream_version = _get_stream_version(tap_stream_id, state)
    state = _set_bk(state, tap_stream_id, "version", stream_version)
    _maybe_activate_before_sync(state, catalog_entry, rep_key, stream_version)
    select, params = _create_sql(catalog_entry, columns, rep_key)
    with get_cursor(config) as cursor:
        LOGGER.info("Running %s PARAMS (%s)", select, params)
        cursor.execute(select, params)
        rows_saved = 0
        for row in cursor:
            rows_saved += 1
            record_message = _row_to_record(catalog_entry, stream_version, row,
                                            columns)
            _emit(record_message)
            if rep_key:
                state = _set_bk(state, tap_stream_id, "replication_key_value",
                                record_message.record[rep_key.column])
            if rows_saved % 1000 == 0:
                _emit(singer.StateMessage(value=state))
        _write_metrics(catalog_entry, rows_saved)
    state = _maybe_activate_after_sync(state, catalog_entry, rep_key, stream_version)
    _emit(singer.StateMessage(value=state))


def sync(config, state, catalog):
    for catalog_entry in catalog.streams:
        state = singer.set_currently_syncing(state, catalog_entry.tap_stream_id)
        _emit(singer.StateMessage(value=state))
        _emit(singer.SchemaMessage(
            stream=catalog_entry.stream,
            schema=catalog_entry.schema.to_dict(),
            key_properties=catalog_entry.key_properties))
        with metrics.job_timer("sync_table") as timer:
            timer.tags["schema"] = catalog_entry.database
            timer.tags["table"] = catalog_entry.table
            _sync_table(config, state, catalog_entry)
    state = singer.set_currently_syncing(state, None)
    _emit(singer.StateMessage(value=state))
